from functools import reduce
from typing import Tuple, List


def and_to_exp(_tuple: Tuple[str, str], eq: bool):
  key, value = _tuple[0], _tuple[1]
  eq_op_sign = "=" if eq else "!="
  return f"{key}{eq_op_sign}{value}"


def or_set_to_exp(_tuple: Tuple[str, List[str]], sign: bool) -> str:
  csv = ', '.join(_tuple[1])
  eq_op = 'in' if sign else 'notin'
  return f"{_tuple[0]} {eq_op} ({csv})"


def ands_to_exps(ands: List[Tuple[str, str]], sign: bool) -> List[str]:
  return [and_to_exp(and_dict, sign) for and_dict in ands]


def ors_to_exp(ors: List[Tuple[str, str]], sign: bool) -> List[str]:
  all_keys = [or_tup[0] for or_tup in ors]

  def agg(whole, key) -> Tuple[str, List[str]]:
    match_tups = [tup for tup in ors if tup[0] == key]
    match_values = [tup[1] for tup in match_tups]
    return whole + [(key, match_values)]

  or_sets = reduce(agg, set(all_keys), [])
  return [or_set_to_exp(or_tup, sign) for or_tup in or_sets]


def assemble_expr_lists(total: List[List[str]]):
  pure = [sub_list for sub_list in total if len(sub_list) > 0]
  macro = [','.join(sub_list) for sub_list in pure]
  return ','.join(macro)


def partition_tups(tups: List[Tuple[str, str]]):
  keys = [t[0] for t in tups]
  repeating_keys = set([k for k in keys if keys.count(k) > 1])
  equalities = [t for t in tups if t[0] not in repeating_keys]
  sets = [t for t in tups if t[0] in repeating_keys]
  return [equalities, sets]


def label_conditions_to_expr(inc_tups, exc_tups) -> str:
  inc_eqs, inc_sets = partition_tups(inc_tups)
  exc_eqs, exc_sets = partition_tups(exc_tups)

  strict_incs = ands_to_exps(inc_eqs, True)
  set_incs = ors_to_exp(inc_sets, True)

  strict_excs = ands_to_exps(exc_eqs, False)
  set_excs = ors_to_exp(exc_sets, False)

  return assemble_expr_lists([
    strict_incs,
    set_incs,
    strict_excs,
    set_excs
  ])




from kubernetes.client import V1Node

from k8kat.auth.kube_broker import broker
from k8kat.res.base.kat_res import KatRes
from k8kat.utils.main import units
from k8kat.utils.main.class_property import classproperty


class KatNode(KatRes):

  @classproperty
  def kind(self):
    return "Node"

  @classmethod
  def is_namespaced(cls) -> bool:
    return False

# --
# --
# --
# -------------------------------INTEL-------------------------------
# --
# --
# --

  def body(self) -> V1Node:
    return self.raw

  def cpu_capacity(self) -> float:
    quant_expr = self.body().status.capacity.get('cpu')
    return units.parse_quant_expr(quant_expr) if quant_expr else None

  def mem_capacity(self) -> float:
    quant_expr = self.body().status.capacity.get('memory')
    return units.parse_quant_expr(quant_expr) if quant_expr else None

  @classmethod
  def k8s_verb_methods(cls):
    return dict(
      read=broker.coreV1.read_node,
      list=broker.coreV1.list_node
    )



from kubernetes.client import V1ObjectMeta, V1ReplicaSetSpec, V1LabelSelector, \
  V1PodTemplateSpec, V1PodSpec, V1Container

from k8kat.auth.kube_broker import broker


def replica_set(**subs):
  default_labels = dict(app=subs['name'])
  labels = {**subs.get('labels', {}), **default_labels}

  return broker.client.V1ReplicaSet(
    metadata=V1ObjectMeta(
      name=subs.get('name'),
      labels=labels
    ),
    spec=V1ReplicaSetSpec(
      replicas=subs.get('replicas', 1),
      selector=V1LabelSelector(
        match_labels=labels
      ),
      template=V1PodTemplateSpec(
        metadata=V1ObjectMeta(
          labels=labels
        ),
        spec=V1PodSpec(
          containers=[
            V1Container(
              name=subs.get('container', 'primary'),
              image=subs.get('image', 'nginx'),
              image_pull_policy="IfNotPresent",
              command=subs.get('command'),
              args=subs.get('args'),
              resources=subs.get('resources', None)
            )
          ]
        )
      )
    )
  )

def create(**subs):
  return broker.appsV1.create_namespaced_replica_set(
    body=replica_set(**subs),
    namespace=subs["ns"]
  )


from typing import Dict, List, Optional

from typing_extensions import TypedDict

from k8kat.auth.kube_broker import broker
from k8kat.utils.main import utils


class ResDefDict(TypedDict):
  name: str
  shortnames: str
  apigroup: str
  namespaced: str
  kind: str


class ApiDefsMan:
  def __init__(self):
    self._defs_list: Optional[List[ResDefDict]] = None

  def defs_list(self, force_reload=False) -> List[ResDefDict]:
    if force_reload or self._defs_list is None:
      self._defs_list = read_defs_list()
    return self._defs_list

  def find_def(self, kind: str) -> Optional[ResDefDict]:
    predicate = lambda e: res_matches_entry(kind, e)
    return next(filter(predicate, self.defs_list()), None)

  def find_api_group(self, kind: str) -> Optional[str]:
    entry = self.find_def(kind)
    return entry and entry['apigroup']

  def kind2plurname(self, kind):
    entry = self.find_def(kind)
    if entry:
      return entry['name']


def read_defs_list() -> List[ResDefDict]:
  kontext = broker.connect_config.get('context')
  raw = utils.k_exec('api-resources', ctx=kontext)
  return raw_defs2dicts(raw)


def raw_defs2dicts(table_str: str) -> List[ResDefDict]:
  lines = table_str.split("\n")
  headers = list(map(str.lower, split_hard(lines[0])))
  header_positions = positions_in_str(lines[0])
  output = []
  for row in lines[1:]:
    row_cells = []
    for header, start_pos in header_positions.items():
      end_pos = row.find(" ", start_pos)
      end_pos = len(row) if end_pos < 0 else end_pos
      row_cells.append(row[start_pos:end_pos].strip())
    entry = {h: row_cells[i] for (i, h) in enumerate(headers)}
    output.append(entry)
  return output


def res_matches_entry(kind: str, entry: ResDefDict) -> bool:
  if kind == entry.get('kind'):
    return True
  if kind.lower() == entry.get('name'):
    return True
  if f"{kind.lower()}s" == entry.get('name'):
    return True
  return False


def positions_in_str(origin: str) -> Dict[str, int]:
  tokens = split_hard(origin)
  return {token: origin.find(token) for token in tokens}


def split_hard(line: str) -> List[str]:
  return ' '.join(line.split()).split(' ')


api_defs_man = ApiDefsMan()



class ClassPropertyDescriptor(object):

  def __init__(self, fget, fset=None):
    self.fget = fget
    self.fset = fset

  def __get__(self, obj, klass=None):
    if klass is None:
      klass = type(obj)
    return self.fget.__get__(obj, klass)()

  def __set__(self, obj, value):
    if not self.fset:
      raise AttributeError("can't set attribute")
    type_ = type(obj)
    return self.fset.__get__(obj, type_)(value)

  def setter(self, func):
    if not isinstance(func, (classmethod, staticmethod)):
      func = classmethod(func)
    self.fset = func
    return self

def classproperty(func):
  if not isinstance(func, (classmethod, staticmethod)):
    func = classmethod(func)

  return ClassPropertyDescriptor(func)



import os
from typing import Optional

from typing_extensions import TypedDict

from k8kat.utils.main import utils

AUTH_TYPE_IN = 'in'
AUTH_TYPE_OUT = 'out'
AUTH_TYPE_KUBE_CONF = 'kube-config'
AUTH_TYPE_SKIP = 'already-authed'


class BrokerConfig(TypedDict, total=False):
  auth_type: str
  sa_name: str
  sa_ns: str
  kubectl: str
  crb_name: str
  cluster_name: str
  context: str


_prod_defaults = BrokerConfig(
  auth_type='in',
)


_dev_defaults = BrokerConfig(
  auth_type=AUTH_TYPE_KUBE_CONF,
  sa_name='nectar-dev',
  sa_ns='default',
  kubectl='kubectl',
  crb_name='nectar',
  cluster_name='dev'
)


_test_defaults = BrokerConfig(
  auth_type=AUTH_TYPE_KUBE_CONF,
)


def env_default_config() -> BrokerConfig:
  if utils.is_prod():
    return _prod_defaults
  elif utils.is_dev():
    return _dev_defaults
  elif utils.is_test():
    return _test_defaults
  else:
    print(f"[broker_configs] WARN unknown env {utils.run_env()}")


def _load_value(env_var_key: str, dict_key: str) -> Optional[str]:
  env = os.environ
  from_env_default = env_default_config().get(dict_key)
  from_specific_env = env.get(f"{utils.run_env().upper()}_{env_var_key}")
  from_global_env = env.get(env_var_key)
  return from_specific_env or from_global_env or from_env_default


def read_env_config() -> BrokerConfig:
  return BrokerConfig(
    auth_type=_load_value('CONNECT_TYPE', 'auth_type'),
    sa_name=_load_value('CONNECT_SA_NAME', 'sa_name'),
    sa_ns=_load_value('CONNECT_SA_NS', 'sa_ns'),
    crb_name=_load_value('CONNECT_CRB_NAME', 'cbr_name'),
    kubectl=_load_value('CONNECT_KUBECTL', 'kubectl'),
    cluster_name=_load_value('CONNECT_CLUSTER', 'cluster_name'),
    context=_load_value('CONNECT_CONTEXT', 'context')
  )




class KatEvent:

  def __init__(self, raw):
    self.raw = raw

  @property
  def name(self):
    return self.raw.metadata.name

  @property
  def namespace(self):
    return self.raw.metadata.namespace

  @property
  def for_uid(self):
    return self.raw.involved_object.uid

  @property
  def for_kind(self):
    return self.raw.involved_object.kind

  @property
  def for_ns(self):
    return self.raw.involved_object.namespace

  @property
  def for_name(self):
    return self.raw.involved_object.name

  @property
  def reason(self):
    return self.raw.reason

  @property
  def message(self):
    return self.raw.message

  @property
  def severity(self):
    return self.raw.type

  @property
  def occurred_at(self):
    return self.raw.first_timestamp

  def meaning(self):
    if self.is_config_map_err():
      return "This suggests it's not reading a ConfigMap right."
    elif self.is_secrets_err():
      return "This suggests it's not reading a Secret right."
    elif self.is_probe_err():
      return "This suggests your liveness probe is failing."
    elif self.is_insufficient_cpu():
      return "This suggests your pod may be CPU starved."
    elif self.is_app_crashing():
      return "This suggests the app inside the container keeps crashing at launch."

  def is_for(self, kat_res):
    return self.for_uid == kat_res.uid

  def is_config_map_err(self):
    return "configmaps" in self.message

  def is_secrets_err(self):
    return "with: secrets" in self.message

  def is_probe_err(self):
    return self.reason == 'Unhealthy'

  def is_app_crashing(self):
    return "Back-off restarting" in self.message

  def is_insufficient_cpu(self):
    return "Insufficient cpu" in self.message



from functools import lru_cache
from typing import Dict, List

from kubernetes.client import V1PodSpec, V1Container, V1Scale, V1ScaleSpec, V1Deployment

from k8kat.auth.kube_broker import broker
from k8kat.res.base.kat_res import KatRes, MetricsDict
from k8kat.res.base.label_set_expressions import label_conditions_to_expr
from k8kat.res.pod.kat_pod import KP
from k8kat.res.relation.relation import Relation
from k8kat.utils.main.class_property import classproperty


class KatDep(KatRes):

  @classproperty
  def kind(self):
    return "Deployment"

  @classproperty
  def kind_aliases(self) -> List[str]:
    return [
      *super().kind_aliases,
      'deployment.apps',
      'deploy'
    ]

  @property
  def pod_spec(self) -> V1PodSpec:
    return self.raw.spec.template.spec

  @property
  def pod_select_labels(self) -> Dict[str, str]:
    return self.raw.spec.selector.match_labels or {}

  @property
  def template_labels(self) -> Dict[str, str]:
    return self.raw.spec.template.metadata.labels or {}

  @property
  def desired_replicas(self):
    return self.raw.spec.replicas

  @property
  def ready_replicas(self):
    return self.raw.status.ready_replicas

# --
# --
# --
# -------------------------------INTEL-------------------------------
# --
# --
# --

  def ternary_status(self):
    if self.is_running_normally():
      return 'positive'
    elif self.has_broken_pod():
      return 'negative'
    else:
      return 'pending'

  def is_running_normally(self):
    return self.ready_replicas == self.desired_replicas

  def has_broken_pod(self) -> bool:
    from k8kat.res.pod.kat_pod import KatPod
    pods: List[KatPod] = self.pods()
    return len([p for p in pods if p.is_broken()]) > 0

  def has_settled(self) -> bool:
    from k8kat.res.pod.kat_pod import KatPod
    pods: List[KatPod] = self.pods()
    pod_settle_states = list(map(KatPod.has_settled, pods))
    return len(pods) == 0 or set(pod_settle_states) == {True}

  @lru_cache(maxsize=128)
  def load_per_pod_metrics(self) -> List[MetricsDict]:
    """Loads the appropriate metrics dict from k8s metrics API."""
    label_sel = label_conditions_to_expr(self.pod_select_labels.items(), [])
    return broker.custom.list_namespaced_custom_object(
      group='metrics.k8s.io',
      version='v1beta1',
      namespace=self.namespace,
      label_selector=label_sel,
      plural='pods'
    )['items']

# --
# --
# --
# -------------------------------ACTION-------------------------------
# --
# --
# --

  def body(self) -> V1Deployment:
    return self.raw

  def container_spec(self, index=0) -> V1Container:
    specs = self.pod_spec.containers
    return specs[index] if len(specs) else None

  def image_name(self, index=0) -> str:
    container_spec = self.container_spec(index)
    return container_spec and container_spec.image

  def container_name(self, index=0) -> str:
    spec = self.container_spec(index)
    return spec.name if spec else None

  def ipp(self, index=0) -> str:
    cont_spec = self.container_spec(index)
    return cont_spec and cont_spec.image_pull_policy

  def replace_image(self, new_image_name):
    self.raw.spec.template.spec.containers[0].image = new_image_name
    self._perform_patch_self()

  def restart_pods(self):
    remember_replicas = self.desired_replicas
    self.scale(0)
    self.scale(remember_replicas)

  def scale(self, replicas):
    broker.appsV1.patch_namespaced_deployment_scale(
      name=self.name,
      namespace=self.ns,
      body=V1Scale(
        spec=V1ScaleSpec(
          replicas=replicas
        )
      )
    )

  @classmethod
  def k8s_verb_methods(cls):
    return dict(
      read=broker.appsV1.read_namespaced_deployment,
      patch=broker.appsV1.patch_namespaced_deployment,
      delete=broker.appsV1.delete_namespaced_deployment,
      list=broker.appsV1.list_namespaced_deployment
    )

# --
# --
# --
# -------------------------------RELATIONS-------------------------------
# --
# --
# --

  def pods(self, **query) -> List[KP]:
    """Selects and returns pods associated with the deployment."""
    from k8kat.res.pod.kat_pod import KatPod
    return Relation[KatPod](
      model_class=KatPod,
      ns=self.ns,
      labels=self.pod_select_labels,
      **query
    )



from typing import Dict

from kubernetes.client import V1ServicePort

from k8kat.auth.kube_broker import broker
from k8kat.res.base.kat_res import KatRes
from k8kat.utils.main import utils
from k8kat.utils.main.class_property import classproperty


class KatSvc(KatRes):

  @classproperty
  def kind(self):
    return "Service"

  @property
  def pod_select_labels(self) -> Dict[str, str]:
    return self.raw.spec.selector or {}

  @property
  def main_port_obj(self) -> V1ServicePort:
    ports = self.raw.spec.ports
    return len(ports) and ports[0]

  @property
  def internal_ip(self) -> str:
    return self.raw.spec.cluster_ip

  @property
  def external_ip(self) -> str:
    load_bal = self.raw.status.load_balancer
    return utils.try_or(lambda: load_bal.ingress[0].ip)

  @property
  def from_port(self) -> int:
    port_obj = self.main_port_obj
    return port_obj and port_obj.port

  @property
  def to_port(self):
    port_obj = self.main_port_obj
    return port_obj and port_obj.target_port

  @property
  def short_dns(self) -> str:
    return f"{self.name}.{self.namespace}"

  @property
  def fqdn(self) -> str:
    return f"{self.short_dns}.svc.cluster.local"

  @property
  def type(self) -> str:
    return self.raw.spec.type

  def raw_endpoints(self):
    return broker.coreV1.read_namespaced_endpoints(self.name, self.ns)

  def flat_endpoints(self):
    raw_endpoints = self.raw_endpoints()
    per_sub = lambda sub: [addr for addr in (sub.addresses or [])]
    return utils.flatten([per_sub(sub) for sub in raw_endpoints.subsets])

  @classmethod
  def k8s_verb_methods(cls):
    return dict(
      read=broker.coreV1.read_namespaced_service,
      patch=broker.coreV1.patch_namespaced_service,
      delete=broker.coreV1.delete_namespaced_service,
      list=broker.coreV1.list_namespaced_service
    )

  @property
  def endpoint_ips(self):
    return [ep.ip for ep in self.flat_endpoints()]



from typing import Callable, TypeVar, List, Dict, Optional, Type

from k8kat.res.base.kat_res import KatRes
from k8kat.res.base.label_set_expressions import label_conditions_to_expr

KT = TypeVar('KT', bound=KatRes)


class Relation(List[KT]):

  def __init__(self, model_class: Type[KT], ns=None, **query):
    if not model_class:
      raise RuntimeError("Cannot compute relation without model_class")

    self.model_class: Type[KT] = model_class
    self.ns: str = ns
    self._query: Dict[str, str] = query

    raw_k8s_res_list = self.perform_in_ns()
    kat_res_list = list(map(model_class, raw_k8s_res_list))
    super().__init__(kat_res_list)

  def perform_in_ns(self):
    impl = self.namespaced_query_impl()
    if self.model_class.is_namespaced():
      return normalize_response(impl(
        namespace=self.ns,
        **self.logical_to_k8s_query()
      ))
    else:
      return normalize_response(impl(
        **self.logical_to_k8s_query()
      ))

  def logical_to_k8s_query(self) -> Dict[str, Optional[str]]:
    k8s_query = dict()
    k8s_query['label_selector'] = process_labels(self._query)
    k8s_query['field_selector'] = process_fields(self._query)
    return k8s_query

  def query(self, **q):
    return Relation(
      self.model_class,
      self.ns,
      **self._query,
      **q
    )

  def namespaced_query_impl(self) -> Callable:
    return self.model_class.k8s_verb_methods()['list']


def process_labels(query) -> Optional[str]:
  with_l, without_l = query.get('labels', {}), query.get('not_labels', {})
  if with_l or without_l:
    return label_conditions_to_expr(with_l.items(), without_l.items())
  else:
    return ''


def process_fields(query) -> Optional[str]:
  with_l, without_l = query.get('fields', {}), query.get('not_fields', {})
  if with_l or without_l:
    return label_conditions_to_expr(with_l.items(), without_l.items())
  else:
    return ''


def normalize_response(result):
  if type(result) == dict:
    return result['items']
  else:
    return result.items



from typing import List

from k8kat.auth.kube_broker import broker
from k8kat.res.node.kat_node import KatNode
from k8kat.utils.main import utils, units
from k8kat.utils.main.types import NodeMetricsDict


class KatCluster:

  @classmethod
  def load_per_node_metrics(cls) -> List[NodeMetricsDict]:
    return broker.custom.list_cluster_custom_object(
      group='metrics.k8s.io',
      version='v1beta1',
      plural='nodes'
    ).get('items', [])

  @classmethod
  def resources_available(cls):
    nodes = KatNode.list()
    per_node_cpu = [n.cpu_capacity() or 0 for n in nodes]
    per_node_mem = [n.mem_capacity() or 0 for n in nodes]
    return sum(per_node_cpu), sum(per_node_mem)

  @classmethod
  def resources_used(cls):
    per_node_metrics = cls.load_per_node_metrics()
    per_node_cpu = [read_node_res_used(m, 'cpu') for m in per_node_metrics]
    per_node_mem = [read_node_res_used(m, 'memory') for m in per_node_metrics]
    return sum(per_node_cpu), sum(per_node_mem)

def read_node_res_used(node_metrics: NodeMetricsDict, resource_type: str):
  metric_deep_key = ('usage', resource_type)
  usage_quant_expr = utils.deep_get(node_metrics, *metric_deep_key)
  quant_bytes = units.parse_quant_expr(usage_quant_expr or '0')
  return quant_bytes or 0



import base64

from kubernetes import config, client
import urllib3
from kubernetes.client import CoreV1Api, AppsV1Api, CustomObjectsApi, ExtensionsV1beta1Api, BatchV1Api

from k8kat.auth.broker_configs import read_env_config, BrokerConfig
from k8kat.auth.broker_configs import AUTH_TYPE_IN, AUTH_TYPE_OUT
from k8kat.auth.broker_configs import AUTH_TYPE_KUBE_CONF, AUTH_TYPE_SKIP
from k8kat.utils.main import utils


class BrokerConnException(Exception):
  def __init__(self, message):
    super().__init__(message)


class KubeBroker:

  def __init__(self):
    self.connect_config = {}
    self.is_connected = False
    self.last_error = None
    self.rbacV1 = None
    self.coreV1 = None
    self.appsV1 = None
    self.client = None
    self.extsV1 = None
    self.custom = None
    self.batchV1 = None

  def connect(self, _config: BrokerConfig = None) -> bool:
    # traceback.print_stack()
    connect_config = _config or read_env_config()
    connect_type = connect_config['auth_type']

    if connect_type == AUTH_TYPE_IN:
      outcome = self.connect_in_cluster()
    elif connect_type == AUTH_TYPE_OUT:
      outcome = self.connect_out_cluster(connect_config)
    elif connect_type == AUTH_TYPE_KUBE_CONF:
      outcome = self.connect_kube_config(connect_config)
    elif connect_type == AUTH_TYPE_SKIP:
      outcome = self.test_connected()
    else:
      outcome = None

    if outcome:
      self.load_api()

    self.is_connected = outcome
    self.connect_config = connect_config
    return outcome

  def test_connected(self) -> bool:
    return self.coreV1.read_namespaced_pod()

  def connect_or_raise(self, passed_config=None):
    if not self.connect(passed_config):
      raise Exception("Cluster authentication failed")

  def load_api(self):
    self.client = client
    self.rbacV1 = client.RbacAuthorizationV1Api()
    self.coreV1: CoreV1Api = client.CoreV1Api()
    self.appsV1: AppsV1Api = client.AppsV1Api()
    self.custom: CustomObjectsApi = client.CustomObjectsApi()
    self.extsV1: ExtensionsV1beta1Api = client.ExtensionsV1beta1Api()
    self.batchV1: BatchV1Api = client.BatchV1Api()

  def connect_in_cluster(self):
    try:
      print(f"[k8kat::kube_broker] In-cluster auth...")
      config.load_incluster_config()
      print(f"[k8kat::kube_broker] In-cluster auth success.")
      return True
    except Exception as e:
      print(f"[kube_broker] In-cluster connect Failed: {e}")
      self.last_error = e
      return False

  def connect_kube_config(self, _config: BrokerConfig):
    try:
      print("[k8kat::kube_broker] Default config auth...")
      config.load_kube_config(context=_config.get('context'))
      print("[k8kat::kube_broker] Default config auth success")
      return True
    except Exception as e:
      print(f"[kube_broker] In-cluster connect Failed: {e}")
      self.last_error = e
      return False

  def connect_out_cluster(self, _config: BrokerConfig):
    rep = _out_conn_debug_strs(_config)
    try:
      print(f"[k8kat::kube_broker] Out-cluster auth ({rep})...")
      configuration = prep_out_client_config(_config)
      client.Configuration.set_default(configuration)
      print(f"[k8kat::kube_broker] Out-cluster auth success")
      return True
    except Exception as e:
      print(f"[k8kat::kube_broker] Out-cluster auth failed {rep}")
      self.last_error = e
      return False

  def is_in_cluster_auth(self):
    return self.connect_config['auth_type'] == 'in'

  def kubectl(self):
    return self.connect_config['kubectl']

  def check_connected_or_raise(self):
    if not self.is_connected:
      if not self.connect():
        raise BrokerConnException(self.last_error or "unknown")


broker = KubeBroker()


def _out_conn_debug_strs(_config: BrokerConfig) -> str:
  context = _config.get('context')
  cluster_name = _config.get('cluster_name')
  sa_name = _config.get('sa_name')
  sa_ns = _config.get('sa_ns')
  return f"cluster={cluster_name}/{context}, perms={sa_ns}/{sa_name}"


def prep_out_client_config(_config: BrokerConfig):
  urllib3.disable_warnings()
  user_token = read_target_cluster_user_token(_config)
  configuration = client.Configuration()
  configuration.host = read_target_cluster_ip(_config)
  configuration.verify_ssl = False
  configuration.debug = False
  configuration.api_key = {"authorization": f"Bearer {user_token}"}
  return configuration


def read_target_cluster_user_token(_config: BrokerConfig) -> str:
  _config = _config
  ctx, sa_name, sa_ns = _config['context'], _config['sa_name'], _config['sa_ns']
  sa_bundle = utils.jk_exec(f"get sa/{sa_name}", ns=sa_ns, ctx=ctx)
  secret_name = sa_bundle['secrets'][0]['name']
  secret_bundle = utils.jk_exec(f"get secret/{secret_name}", ns=sa_ns, ctx=ctx)
  b64_user_token = secret_bundle['data']['token']
  out = str(base64.b64decode(b64_user_token))[2:-1]
  return out


def read_target_cluster_ip(_config: BrokerConfig) -> str:
  on_board_config = utils.jk_exec('config view')
  clusters = on_board_config['clusters']
  target = _config['cluster_name']
  cluster_bundle = [c for c in clusters if c['name'] == target][0]
  return cluster_bundle['cluster']['server']



import time

from kubernetes.client import V1ObjectMeta, V1PodSpec, V1Container, V1DeploymentSpec, V1PodTemplateSpec, V1LabelSelector

from k8kat.auth.kube_broker import broker

def create(**subs):
  default_labels = dict(app=subs['name'])
  labels = {**subs.get('labels', {}), **default_labels}
  match_labels = {**labels, **subs.get('selector', {})}

  deployment = broker.client.V1Deployment(
    metadata=V1ObjectMeta(
      name=subs['name'],
      labels=labels,
      annotations=subs.get('annotations', {})
    ),
    spec=V1DeploymentSpec(
      replicas=subs.get('replicas', 1),
      selector=V1LabelSelector(
        match_labels=match_labels
      ),
      template=V1PodTemplateSpec(
        metadata=V1ObjectMeta(labels=labels),
        spec=V1PodSpec(
          containers=[
            V1Container(
              name=subs.get("container", "primary"),
              image=subs.get('image', 'nginx'),
              image_pull_policy=subs.get('ipp', 'IfNotPresent'),
              resources=subs.get('resources', None)
            )
          ]
        )
      )
    )
  )

  result = broker.appsV1.create_namespaced_deployment(
    body=deployment,
    namespace=subs['ns']
  )

  eq_exprs = [f"{t[0]}={t[1]}" for t in list(labels.items())]
  def pods():
    return broker.coreV1.list_namespaced_pod(
      namespace=subs['ns'],
      label_selector=",".join(eq_exprs)
    ).items

  while not len(pods()) == deployment.spec.replicas:
    time.sleep(1)

  return result



import os
from typing import List, Tuple

from k8kat.auth.broker_configs import read_env_config
from k8kat.auth.kube_broker import broker
from k8kat.res.pod.kat_pod import KatPod
from k8kat.utils.main import utils
from k8kat.utils.testing import simple_pod, simple_svc, simple_dep

NAMESPACES = ['n1', 'n2', 'n3']


def create_svc(ns: str, name: str, **subs):
  return simple_svc.create(name=name, ns=ns, **subs)


def create_pod(ns: str, name: str, **subs) -> KatPod:
  return simple_pod.create(name=name, ns=ns, **subs)


def nk_label_dep(ns: str, name: str, labels: List[Tuple[str, str]]):
  api = broker.appsV1
  dep = api.read_namespaced_deployment(namespace=ns, name=name)
  dep.metadata.labels = {t[0]: t[1] for t in labels}
  api.patch_namespaced_deployment(namespace=ns, name=name, body=dep)


def k_apply(filename, **kwargs):
  config = read_env_config()
  root = utils.root_path()
  filename = os.path.join(root, f"utils/testing/fixtures/{filename}.yaml")
  kubectl, context = config['kubectl'], config['context']
  utils.k_exec(f"apply -f {filename}", k=kubectl, ctx=context, **kwargs)



import setuptools

with open("README.md", "r") as fh:
  long_description = fh.read()

setuptools.setup(
  name="k8kat",
  version="0.0.211",
  author="Nectar Cloud Software",
  author_email="xavier@codenectar.com",
  description="Sugar for Kubernetes",
  long_description=long_description,
  long_description_content_type="text/markdown",
  url="https://github.com/nectar-cs/k8kat",
  packages=setuptools.find_packages(exclude=[
    "k8kat.tests.*", "k8kat.tests"
  ]),
  install_requires=[
    "python-dotenv",
    "kubernetes",
    "typing-extensions",
    "inflection"
  ],
  classifiers=[
    "Programming Language :: Python :: 3",
    "License :: OSI Approved :: MIT License",
    "Operating System :: OS Independent",
  ],
  python_requires='>=3.6'
)



from typing import Optional, Dict

import yaml
from kubernetes.client import V1ConfigMap

from k8kat.auth.kube_broker import broker
from k8kat.res.base.kat_res import KatRes
import json

from k8kat.utils.main.class_property import classproperty


class KatMap(KatRes):

  @classproperty
  def kind(self):
    return "ConfigMap"

  def body(self) -> V1ConfigMap:
    return self.raw

  @property
  def data(self):
    return self.raw.data

  def jget(self, key=None, backup=None) -> Optional[Dict[str, any]]:
    key = key or 'master'
    raw_value = self.data.get(key)
    obtained = raw_value and json.loads(raw_value)
    return obtained if obtained is not None else backup

  def yget(self, key=None, backup=None) -> Optional[Dict[str, any]]:
    key = key or 'master'
    raw_value = self.data.get(key)
    obtained = raw_value and yaml.load(raw_value, Loader=yaml.FullLoader)
    return obtained if obtained is not None else backup

  def jpatch(self, content: Dict, key: str = None, merge: bool = False):
    key = key or 'master'
    content = {**self.jget(key), **content} if merge else content
    self.raw.data = ({key: json.dumps(content)})
    return self.patch()

  @classmethod
  def k8s_verb_methods(cls):
    return dict(
      read=broker.coreV1.read_namespaced_config_map,
      patch=broker.coreV1.patch_namespaced_config_map,
      delete=broker.coreV1.delete_namespaced_config_map,
      list=broker.coreV1.list_namespaced_config_map
    )



from urllib.parse import unquote_plus
from http.client import HTTPResponse
from io import BytesIO
from typing import List, Optional, Dict, Callable
from kubernetes.client import V1Container

from k8kat.utils.main import units


class FakeSocket:
  def __init__(self, response_bytes):
    self._file = BytesIO(response_bytes)

  def makefile(self, *args, **kwargs):
    return self._file

def coerce_cmd_format(cmd) -> List[str]:
  if isinstance(cmd, str):
    parts = cmd.split(" ")
    # parts = [part.replace("SPACE_CHAR", " ") for part in parts]
    parts = [unquote_plus(part) for part in parts]
    return parts
  else:
    return cmd

def container_req_or_lim(container: V1Container, metric: str, resource: str) -> Optional[float]:
  """Gets container capacity and returns in cores (cpu) / bytes (memory)."""
  assert metric in ['requests', 'limits']
  assert resource in ['cpu', 'memory', 'ephemeral-storage']
  metrics_dict: Dict = getattr(container.resources, metric, None)
  capacity_expr = (metrics_dict or {}).get(resource, None)
  return capacity_expr and units.parse_quant_expr(capacity_expr)


def build_curl_cmd(**params) -> List[str]:
  raw_headers = params.get('headers', {})
  headers = [f"{0}: {1}".format(k, v) for k, v in raw_headers]
  body = params.get('body', None)

  cmd = [
    "curl" if params.get('with_command') else None,
    "-s",
    "-i",
    '-X', params.get('verb', 'GET'),
    '-H' if headers else None,
    headers if headers else None,
    '-d' if body else None, body if body else None,
    "--connect-timeout", "1",
    f"{params['url']}{params.get('path', '/')}"
  ]
  return [part for part in cmd if part is not None]


def parse_response(response_str) -> Optional[HTTPResponse]:
  if response_str:
    source = FakeSocket(response_str.encode('ascii'))
    # noinspection PyTypeChecker
    response = HTTPResponse(source)
    response.begin()
    return response

  else:
    return None


def when_running_normally(func: Callable) -> Callable:
  """Note: needs to be defined at top of class."""
  def running_normally_func(*args, **kwargs):
    if args[0].is_running_normally():
      return func(*args, **kwargs)
    else:
      return None

  return running_normally_func
